//------------------------------------------------------------------------------
// File: PushSourceBitmap.cpp
//
// Desc: DirectShow sample code - In-memory push mode source filter
//       Provides a static bitmap as the video output stream.
//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------------------------

#include <streams.h>

#include "PushSource.h"
#include "PushGuids.h"

/**********************************************
 *
 *  CPushPinTeletext Class
 *  
 *
 **********************************************/

CPushPinTeletext::CPushPinTeletext(HRESULT *phr, CSource *pFilter)
	: CSourceStream(NAME("Push Source Teletext"), phr, pFilter, L"Out")
	, m_hFile(INVALID_HANDLE_VALUE)
	, m_nDataSize(0)
	, m_iFrameNumber(0)
	, m_rtFrameLength(FPS_25) // Display 25 bitmap frames per second
{
    m_hFile = CreateFileA(g_SourceFilePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (m_hFile == INVALID_HANDLE_VALUE) {
		char szMsg[MAX_PATH + MAX_PATH + 100];
		DWORD err = GetLastError();
		StringCchPrintfA(szMsg, _countof(szMsg), "Could not open source file '%s', ERROR=%d.", g_SourceFilePath, err);
		OutputDebugStringA(szMsg);
		MessageBoxA(NULL, szMsg, "PushSourceTeletext filter error", MB_ICONERROR | MB_OK);
		*phr = HRESULT_FROM_WIN32(err);
		return;
    }

	// read first data record
	BYTE tmp[8];
	DWORD nBytesRead = 0;
	if (!ReadFile(m_hFile, tmp, 8, &nBytesRead, NULL)) {
		char szMsg[MAX_PATH + MAX_PATH + 100];
		DWORD err = GetLastError();
		StringCchPrintfA(szMsg, _countof(szMsg), "Could not read file '%s', ERROR=%d.", g_SourceFilePath, err);
		OutputDebugStringA(szMsg);
		MessageBoxA(NULL, szMsg, "PushSourceTeletext filter error", MB_ICONERROR | MB_OK);
		*phr = HRESULT_FROM_WIN32(err);
		return;
	}
	if (tmp[0] != 'd' || tmp[1] != 'a' || tmp[2] != 't' || tmp[3] != 'a') {
		char szMsg[MAX_PATH + MAX_PATH + 100];
		StringCchPrintfA(szMsg, _countof(szMsg), "Wrang format of source file '%s'", g_SourceFilePath);
		OutputDebugStringA(szMsg);
		MessageBoxA(NULL, szMsg, "PushSourceTeletext filter error", MB_ICONERROR | MB_OK);
		*phr = HRESULT_FROM_WIN32(ERROR_BAD_FORMAT);
		return;
	}
	m_nDataSize = (*(DWORD*)&tmp[4]) * 5;
	// reset file pos
	SetFilePointer(m_hFile, 0, NULL, FILE_BEGIN);
}


CPushPinTeletext::~CPushPinTeletext()
{   
    DbgLog((LOG_TRACE, 3, TEXT("Frames written %d"),m_iFrameNumber));

	if (INVALID_HANDLE_VALUE != m_hFile) { // Close and invalidate the file handle, since we have copied its bitmap data
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}
}


// GetMediaType: This method tells the downstream pin what types we support.
HRESULT CPushPinTeletext::GetMediaType(CMediaType *pMediaType)
{
    CAutoLock cAutoLock(m_pFilter->pStateLock());

    CheckPointer(pMediaType, E_POINTER);

	pMediaType->SetType(&MEDIATYPE_Stream);
	pMediaType->SetSubtype(&MEDIATYPE_DVB_TELETEXT);
	pMediaType->SetSampleSize(m_nDataSize);

    return S_OK;
}


HRESULT CPushPinTeletext::DecideBufferSize(IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pRequest)
{
    HRESULT hr;
    CAutoLock cAutoLock(m_pFilter->pStateLock());

    CheckPointer(pAlloc, E_POINTER);
    CheckPointer(pRequest, E_POINTER);

    // If the bitmap file was not loaded, just fail here.
    if (0 == m_nDataSize)
        return E_FAIL;

    // Ensure a minimum number of buffers
    if (pRequest->cBuffers == 0)
    {
        pRequest->cBuffers = 2;
    }
    pRequest->cbBuffer = m_nDataSize;

    ALLOCATOR_PROPERTIES Actual;
    hr = pAlloc->SetProperties(pRequest, &Actual);
    if (FAILED(hr)) 
    {
        return hr;
    }

    // Is this allocator unsuitable?
    if (Actual.cbBuffer < pRequest->cbBuffer) 
    {
        return E_FAIL;
    }

    return S_OK;
}


// This is where we insert the data bits into the stream.
// FillBuffer is called once for every sample in the stream.
HRESULT CPushPinTeletext::FillBuffer(IMediaSample *pSample)
{
    CheckPointer(pSample, E_POINTER);

    // If the bitmap file was not loaded, just fail here.
    if (0 == m_nDataSize)
        return E_FAIL;

    CAutoLock cAutoLockShared(&m_cSharedState);

    // Access the sample's data buffer
	BYTE *pData;
	pSample->GetPointer(&pData);
	long cbData = pSample->GetSize();

	// read data
	BYTE tmp[8];
	DWORD nBytesRead = 0;
	if (!ReadFile(m_hFile, tmp, 8, &nBytesRead, NULL) || nBytesRead != 8) { // no more data
		return S_FALSE;
	}
	ASSERT(tmp[0] == 'd' && tmp[1] == 'a' && tmp[2] == 't' && tmp[3] == 'a');
	DWORD ln = *(DWORD*)&tmp[4];
	ASSERT(ln <= cbData);
	DWORD read_size, skip;
	if (ln <= (DWORD)cbData) {
		read_size = ln;
		skip = 0;
	}
	else {
		read_size = cbData;
		skip = ln - cbData;
	}
	if (!ReadFile(m_hFile, pData, read_size, &nBytesRead, NULL) || nBytesRead < read_size) { // no data
		return S_FALSE;
	}
	if (skip > 0)
		SetFilePointer(m_hFile, skip, NULL, FILE_CURRENT);

	pSample->SetActualDataLength(read_size);

    // Set the timestamps that will govern playback frame rate.
    REFERENCE_TIME rtStart = m_iFrameNumber * m_rtFrameLength;
    REFERENCE_TIME rtStop  = rtStart + m_rtFrameLength;
    pSample->SetTime(&rtStart, &rtStop);
    m_iFrameNumber++;

    // Set TRUE on every sample
    pSample->SetSyncPoint(TRUE);

    return S_OK;
}


/**********************************************
 *
 *  CPushSourceTeletext Class
 *
 **********************************************/

CPushSourceTeletext::CPushSourceTeletext(IUnknown *pUnk, HRESULT *phr)
           : CSource(NAME("PushSourceTeletext"), pUnk, CLSID_PushSourceTeletext)
{
    // The pin magically adds itself to our pin array.
    m_pPin = new CPushPinTeletext(phr, this);

    if (phr)
    {
        if (m_pPin == NULL)
            *phr = E_OUTOFMEMORY;
        else
            *phr = S_OK;
    }  
}


CPushSourceTeletext::~CPushSourceTeletext()
{
    delete m_pPin;
}


CUnknown * WINAPI CPushSourceTeletext::CreateInstance(IUnknown *pUnk, HRESULT *phr)
{
	CPushSourceTeletext *pNewFilter = new CPushSourceTeletext(pUnk, phr );

    if (phr)
    {
        if (pNewFilter == NULL) 
            *phr = E_OUTOFMEMORY;
        else
            *phr = S_OK;
    }

    return pNewFilter;
}
